/*-
 * Copyright (c) 2010 Michael Schnell
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <libutil.h>
#include <signal.h>

/* Defines */
#define EMERGENCY	5
#define TZ_ZEROC        2732
#define DEFAULT_POLL	500000

#define CK(temp) temp * 10 + TZ_ZEROC

/* Prototypes */
static void signal_handler(int);

/* Global variables */
int sig_int = 0;
int sig_term = 0;

int
main(int argc, char *argv[])
{
	int ch, v_flag, i, mib[4], mib_fanlevel[4];
	int fanlevel, lastfanlevel, fancontrol, poll_val;
	int temp, tempi, lasttemp, cpus;
	struct pidfh *pfh = NULL;
	struct timeval timeout;
	size_t len_int, len_temp;
	char temp_string[128];
	int **smib;
	pid_t pid;

	if (geteuid() != 0) {
		errx(1, "you are not root");
	}

	/* Set default polling interval */
	poll_val = DEFAULT_POLL;

	v_flag = 0;

	/* Arguments */
	while ((ch = getopt(argc, argv, "i:v")) != -1) {
		switch (ch) {
		case 'i':
			poll_val = atoi(optarg);
			if (poll_val < 100 || poll_val > 5000000) {
				errx(1, "not allowed polling interval");
			}
			break;
		case 'v':
			v_flag = 1;
			break;
		default:
			errx(1, "wrong parameters");
		}
	}

	if (v_flag) {
		printf("Polling intervall is: %ims\n", poll_val);
	}

	/* run in daemon mode */
	if (!v_flag) {
		pfh = pidfile_open("/var/run/loolerd.pid", 0600, &pid);
		if (pfh == NULL) {
			if (errno == EEXIST) {
				errx(1, "pidfile already exists with PID: %d", pid);
			}

			warn("cannot open pidfile");
		}

		if (daemon(0, 0) !=0) {
			warn("cannot enter daemon mode");
			pidfile_remove(pfh);

			return EXIT_FAILURE;
		}
	
		pidfile_write(pfh);
	}

	/* Signal handler */
	signal(SIGINT, signal_handler);
	signal(SIGTERM, signal_handler);

	len_int = sizeof(int); /* size of an integer */
	len_temp = sizeof(int); /* size of an integer for temp */

	/* Get CPUs */
	if (sysctlbyname("kern.smp.cpus", &cpus, &len_int, NULL, 0) != 0) {
		errx(1, "can't get number of CPUs");
	}

	/* Allocate space for mibs */
	if ((smib = malloc(cpus * sizeof(int *))) == NULL) {
		errx(1, "malloc for MIB structure failed");
	}

	/* Allocate space and get mib for each CPU in order to achieve fast
	 * sysctl requests in the main loop.
	 */
	for (i = 0; i < cpus; ++i) {
		if ((smib[i] = malloc(4 * sizeof(int))) == NULL) {
			errx(1, "malloc for MIB substructure failed");
		}

		sprintf(temp_string, "dev.cpu.%i.temperature", i);
		
	 	if (sysctlnametomib(temp_string, smib[i], &len_int)) {
			err(1, "can't lookup temperature");
		}
	}

	/* Create mib for fast fan level sysctl access */
	if (sysctlnametomib("dev.acpi_ibm.0.fan_level", mib_fanlevel, &len_int)) {
		err(1, "can't lookup fan level");
	}
	
	/* Take over cooler control */
	fancontrol = 0;

	if (sysctlbyname("dev.acpi_ibm.0.fan", NULL, NULL, &fancontrol, len_int) != 0) {
		errx(1, "can't access IBM ACPI sysctl");
	}

	/* main loop */
	while (1) {
		/* If receive a signal break the loop */
		if (sig_int || sig_term) {
			break;
		}

		/* Find hottest CPU/Core */
		for (i = 0; i < cpus; ++i) {
			temp = 0; /* reset max temperature */

			if (sysctl(smib[i], 4, &tempi, &len_temp, NULL, 0) != 0) {
				errx(1, "can't find temp");
			}

			if (tempi > temp) {
				temp = tempi;
			}
		}

		/* Emergency start */
		if (temp > CK(64)) {
			fanlevel = 7;

			if (sysctl(mib_fanlevel, 4, NULL, 0, &fanlevel, len_int) != 0) {
				errx(1, "can't set fan_level");
			}

			if (v_flag) {
				printf("Emergency startup\n");
			}

			/* Use select for precise timeouts */
			timeout.tv_sec = EMERGENCY;
			timeout.tv_usec = 0;

			select(0, NULL, NULL, NULL, &timeout);

			continue;
		}

		/* Fan Levels: 	0 	off
		 * 		1	~2000/rpm
		 * 		2	~3500/rpm	
		 * 		3	~3700/rpm
		 * 		4	same
		 * 		5	~3850/rpm
		 * 		6	same
		 * 		7	~4500/rpm
		 */
		if (temp < CK(37)) {
			fanlevel = 1;
		}
		
		if (temp >= CK(37)) {
			fanlevel = 2;
		}
		
		if (temp >= CK(45)) {
			fanlevel = 3;
		}
		
		if (temp >= CK(53)) {
			fanlevel = 5;
		} 
		
		if (temp >= CK(58)) {
			fanlevel = 7;
		}

		/* Write sysctl if it has changed */
		if (lastfanlevel != fanlevel) {
			if (sysctl(mib_fanlevel, 4, NULL, 0, &fanlevel, len_int) != 0) {
				errx(1, "can't set fan_level");
			}
		}

		lastfanlevel = fanlevel;
		lasttemp = temp;
	
		if (v_flag) {
			printf("Temp: %i Fanlevel: %i\n", (temp - TZ_ZEROC)/10, fanlevel);
		}

		/* Use select for presize timeouts */
		timeout.tv_sec = poll_val / 1000000;
		timeout.tv_usec = poll_val % 1000000;

		select(0, NULL, NULL, NULL, &timeout);
	}

	/* Restore hardware cooler control */
	fancontrol = 1;

	if (sysctlbyname("dev.acpi_ibm.0.fan", NULL, NULL, &fancontrol, len_int) != 0) {
		errx(1, "can't restore sysctl");
	}

	if (!v_flag) {
		pidfile_remove(pfh);
	}

	/* Free allocated variables */
	for (i = 0; i < cpus; ++i) {
		free(smib[i]);
	}

	free(smib);

	return EXIT_SUCCESS;
}

/*
 * The signale handler function, which simply sets a global variable, so
 * the main program can act on it accordingly.
 */
static void
signal_handler(int sig)
{
	switch(sig) {
		case SIGINT:
			sig_int = 1;
			break;
		case SIGTERM:
			sig_term = 1;
			break;
	}
}

